'''
static_result:
submodule for reporting fitting process of static spectrum result

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Callable
import numpy as np
from scipy.special import wofz
import h5py as h5
from ..mathfun.peak_shape import voigt, \
    edge_gaussian, edge_lorenzian


class StaticResult(dict):
    '''
    Represent results for fitting static driver routine

    Attributes:
     model ({'thy', 'voigt'}): model used for fitting

            `voigt`: sum of voigt function, edge function and base funtion

            `thy`: sum of voigt broadened theoretical lineshape spectrum, edge function and base function
     thy_peak (sequence of np.ndarray): theoretical calculated peak position and intensity [model: `thy`]
     policy ({'shift', 'scale', 'both'}): policy to match discrepancy between theoretical spectrum and
      experimental static spectrum.
     e (np.ndarray): energy range
     intensity (np.ndarray): intensity of static spectrum
     eps (np.ndarray): estimated error of static spectrum
     fit (np.ndarray): fitting curve for data (n,)
     fit_comp (np.ndarray): curve for each voigt component and edge
     base (np.ndaray): fitting curve for baseline
     res (np.ndarray): residual curve (data-fit) for static spectrum (n,)
     edge ({'g', 'l'}): type of edge function, if edge is None then edge function is not
      included in the fitting model

          'g': gaussian type edge function

          'l': lorenzian type edge function
     base_order (int): order of baseline function
                       if base_order is None then baseline is not included in the fitting model
     param_name (np.ndarray): name of parameter
     n_voigt (int): number of voigt component
     n_edge (int): number of edge component
     x (np.ndarray): best parameter
     bounds (sequence of tuple): boundary of each parameter
     c (np.ndarray): best weight of each voigt component and edge of data
     chi2 (float): chi squared value of fitting
     aic (float): Akaike Information Criterion statistic: :math:`N\\log(\\chi^2/N)+2N_{parm}`
     bic (float): Bayesian Information Criterion statistic: :math:`N\\log(\\chi^2/N)+N_{parm}\\log(N)`
     red_chi2 (float): total reduced chi squared value of fitting
     nfev (int): total number of function evaluation
     n_param (int): total number of effective parameter
     num_pts (int): total number of data points
     jac (np.ndarray): jacobian of objective function at optimal point
     cov (np.ndarray): covariance matrix (i.e. inverse of :math:`J^T J`)
     cov_scaled (np.ndarray): scaled covariance matrix (i.e. :math:`\\chi^2_{red} \\cdot {cov}`)
     corr (np.ndarray): parameter correlation matrix
     x_eps (np.ndarray): estimated error of parameter
      (i.e. square root of diagonal element of `conv_scaled`)
     method_glb ({'ampgo', 'basinhopping'}):
      method of global optimization used in fitting process
     message_glb (str): messages from global optimization process
     method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                           minimization (refinement of global optimization solution)
     success_lsq (bool): whether or not local least square optimization is successed
     message_lsq (str): messages from local least square optimization process
     status ({0, -1}): status of optimization process

                   `0` : normal termination

                   `-1` : least square optimization process is failed
    '''

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __dir__(self):
        return list(self.keys())

    def __str__(self, corr_tol: float = 1e-1):
        '''
        Report StaticResult
        Args:
        corr_tol: parameter correlation greather than `corr_tol` would be reported
        '''
        doc_lst = []
        doc_lst.append('[Model information]')
        doc_lst.append(f"    model : {self['model']}")
        if self['model'] == 'thy':
            doc_lst.append(f"    policy: {self['policy']}")
        if self['edge'] is not None:
            doc_lst.append(f"    edge: {self['edge']}")
        if self['base_order'] is not None:
            doc_lst.append(f"    base_order: {self['base_order']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Method]')
        if self['method_glb'] is not None:
            doc_lst.append(f"    global: {self['method_glb']}")
        doc_lst.append(f"    leastsq: {self['method_lsq']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Status]')
        doc_lst.append(f"    nfev: {self['nfev']}")
        doc_lst.append(f"    status: {self['status']}")
        if self['method_glb'] is not None:
            doc_lst.append(f"    global_opt msg: {self['message_glb']}")
        doc_lst.append(f"    leastsq_opt msg: {self['message_lsq']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Results]')
        doc_lst.append(f"    Data points: {self['num_pts']}")
        doc_lst.append(
            f"    Number of effective parameters: {self['n_param']}")
        doc_lst.append(
            f"    Degree of Freedom: {self['num_pts']-self['n_param']}")
        doc_lst.append(
            f"    Chi squared: {self['chi2']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    Reduced chi squared: {self['red_chi2']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    AIC (Akaike Information Criterion statistic): {self['aic']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    BIC (Bayesian Information Criterion statistic): {self['bic']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(' ')
        doc_lst.append('[Parameters]')
        for pn, pv, p_err in zip(self['param_name'], self['x'], self['x_eps']):
            doc_lst.append(
                f'    {pn}: {pv: .8f} +/- {p_err: .8f} ({100*np.abs(p_err/pv): .2f}%)'.rstrip('0').rstrip('.'))
        doc_lst.append(' ')
        doc_lst.append('[Parameter Bound]')
        for pn, pv, bd in zip(self['param_name'], self['x'], self['bounds']):
            doc_lst.append(f'    {pn}: {bd[0]: .8f}'.rstrip('0').rstrip(
                '.')+f' <= {pv: .8f} <= {bd[1]: .8f}'.rstrip('0').rstrip('.'))
        doc_lst.append(' ')

        doc_lst.append('[Component Contribution]')
        doc_lst.append('    Static spectrum')
        if self['base_order'] is None:
            coeff_abs = np.abs(self['c'])
            coeff_contrib = 100*self['c']
        else:
            coeff_abs = np.abs(self['c'][:-self['base_order']-1])
            coeff_contrib = 100*self['c'][:-self['base_order']-1]
        coeff_sum = np.sum(coeff_abs)
        coeff_contrib = coeff_contrib/coeff_sum
        for v in range(self['n_voigt']):
            row = [f"     {self['model']} {v+1}:"]
            row.append(f'{coeff_contrib[v]: .2f}%')
            doc_lst.append(' '.join(row))
        if self['edge'] is not None:
            for e in range(self['n_edge']):
                row = [f"     {self['edge']} type edge {e+1}:"]
                row.append(f"{coeff_contrib[self['n_voigt']+e]: .2f}%")
                doc_lst.append(' '.join(row))
        doc_lst.append(' ')

        doc_lst.append('[Parameter Correlation]')
        doc_lst.append(f'    Parameter Correlations > {corr_tol: .3f}'.rstrip(
            '0').rstrip('.') + ' are reported.')

        A = np.empty((len(self['x']), len(self['x'])), dtype=object)
        for i in range(len(self['x'])):
            for j in range(len(self['x'])):
                A[i, j] = (i, j)

        mask = (np.abs(self['corr']) > corr_tol)

        for pair in A[mask]:
            if pair[0] > pair[1]:
                tmp_str_lst = [f"    ({self['param_name'][pair[0]]},"]
                tmp_str_lst.append(f"{self['param_name'][pair[1]]})")
                tmp_str_lst.append('=')
                tmp_str_lst.append(
                    f"{self['corr'][pair]: .3f}".rstrip('0').rstrip('.'))
                doc_lst.append(' '.join(tmp_str_lst))

        return '\n'.join(doc_lst)


def save_StaticResult(result: StaticResult, filename: str):
    '''
    save static fitting result to the h5 file

    Args:
     result: static fitting result
     filename: filename to store result. It will store result to filename.h5

    Returns:
     h5 file which stores result
    '''
    model_key_lst = ['chi2', 'n_param', 'num_pts', 'red_chi2',
                     'aic', 'bic', 'nfev', 'method_lsq', 'success_lsq', 'message_lsq', 'status']

    with h5.File(f'{filename}.h5', 'w') as f:
        expt = f.create_group('experiment')
        fit_res = f.create_group('fitting_result')
        if result['model'] == 'thy':
            thy_stick = f.create_group('theoretical_peaks')
            for i in range(result['n_voigt']):
                thy_stick.create_dataset(
                    f'species {i+1}', data=result['thy_peak'][i])
            fit_res.attrs['policy'] = result['policy']
        expt.create_dataset('energy', data=result['e'])
        expt.create_dataset('intensity', data=result['intensity'])
        expt.create_dataset('error', data=result['eps'])
        fit_res.attrs['model'] = result['model']
        fit_res.attrs['n_voigt'] = result['n_voigt']
        fit_res.attrs['n_edge'] = result['n_edge']
        if result['edge'] is not None:
            fit_res.attrs['edge'] = result['edge']
        if result['base_order'] is not None:
            fit_res.attrs['base_order'] = result['base_order']
        else:
            fit_res.attrs['base_order'] = 'no'
        if result['method_glb'] is not None:
            fit_res.attrs['method_glb'] = result['method_glb']
            fit_res.attrs['message_glb'] = result['message_glb']
        else:
            fit_res.attrs['method_glb'] = 'no'

        for k in model_key_lst:
            fit_res.attrs[k] = result[k]

        fit_res_curve = fit_res.create_group('fit_curve')
        fit_res_curve.create_dataset('fit', data=result['fit'])
        fit_res_curve.create_dataset('fit_comp', data=result['fit_comp'])
        fit_res_curve.create_dataset('weight', data=result['c'])
        if result['base_order'] is not None:
            fit_res_curve.create_dataset('base', data=result['base'])
        fit_res_curve.create_dataset('res', data=result['res'])
        fit_res_param = fit_res.create_group('parameter')
        fit_res_param.create_dataset('param_opt', data=result['x'])
        fit_res_param.create_dataset('param_eps', data=result['x_eps'])
        fit_res_param.create_dataset(
            'param_name',
            data=np.char.encode(result['param_name'].astype('str'),
            encoding='utf-8').astype('S100'),
            dtype='S100')
        fit_res_param.create_dataset('param_bounds', data=result['bounds'])
        fit_res_param.create_dataset('correlation', data=result['corr'])
        fit_res_mis = fit_res.create_group('miscellaneous')
        fit_res_mis.create_dataset('jac', data=result['jac'])
        fit_res_mis.create_dataset('cov', data=result['cov'])
        fit_res_mis.create_dataset('cov_scaled', data=result['cov_scaled'])


def load_StaticResult(filename: str) -> StaticResult:
    '''
    load static fitting result from h5 file

    Args:
     filename: filename to load result. It will load result to filename.h5

    Returns:
     loaded StaticResult instance
    '''
    model_key_lst = ['chi2', 'n_param', 'num_pts', 'red_chi2',
                     'aic', 'bic', 'nfev',
                     'method_lsq', 'success_lsq', 'message_lsq', 'status']

    result = StaticResult()

    with h5.File(f'{filename}.h5', 'r') as f:
        expt = f['experiment']
        fit_res = f['fitting_result']
        result['e'] = np.atleast_1d(expt['energy'])
        result['intensity'] = np.atleast_1d(expt['intensity'])
        result['eps'] = np.atleast_1d(expt['error'])
        result['model'] = fit_res.attrs['model']
        result['n_voigt'] = fit_res.attrs['n_voigt']
        result['n_edge'] = fit_res.attrs['n_edge']
        if result['model'] == 'thy':
            thy_stick = f['theoretical_peaks']
            result['thy_peak'] = np.empty(result['n_voigt'], dtype=object)
            for i in range(result['n_voigt']):
                result['thy_peak'][i] = \
                    np.atleast_2d(thy_stick[f'species {i+1}'])
            result['policy'] = fit_res.attrs['policy']
        if fit_res.attrs['n_edge'] != 0:
            result['edge'] = fit_res.attrs['edge']
        else:
            result['edge'] = None
        if fit_res.attrs['base_order'] != 'no':
            result['base_order'] = fit_res.attrs['base_order']
        else:
            result['base_order'] = None

        if fit_res.attrs['method_glb'] == 'no':
            result['method_glb'] = None
            result['message_glb'] = None
        else:
            result['method_glb'] = fit_res.attrs['method_glb']
            result['message_glb'] = fit_res.attrs['message_glb']

        for k in model_key_lst:
            result[k] = fit_res.attrs[k]
        fit_res_curve = fit_res['fit_curve']

        result['fit'] = np.atleast_1d(fit_res_curve['fit'])
        result['fit_comp'] = np.atleast_1d(fit_res_curve['fit_comp'])
        result['c'] = np.atleast_1d(fit_res_curve['weight'])
        if result['base_order'] is not None:
            result['base'] = np.atleast_1d(fit_res_curve['base'])
        else:
            result['base'] = None
        result['res'] = np.atleast_1d(fit_res_curve['res'])

        fit_res_param = fit_res['parameter']
        result['x'] = np.atleast_1d(fit_res_param['param_opt'])
        result['x_eps'] = np.atleast_1d(fit_res_param['param_eps'])
        result['param_name'] = \
            np.char.decode(np.atleast_1d(fit_res_param['param_name']),
            encoding='utf-8')
        tmp = np.atleast_2d(fit_res_param['param_bounds'])
        lst = tmp.shape[0]*[None]
        for i in range(tmp.shape[0]):
            lst[i] = (tmp[i, 0], tmp[i, 1])
        result['bounds'] = lst
        result['corr'] = np.atleast_2d(fit_res_param['correlation'])

        fit_res_mis = fit_res['miscellaneous']
        result['jac'] = np.atleast_2d(fit_res_mis['jac'])
        result['cov'] = np.atleast_2d(fit_res_mis['cov'])
        result['cov_scaled'] = np.atleast_2d(fit_res_mis['cov_scaled'])
    return result


def deriv_edge_g_aux(e: np.ndarray, fwhm_G: float) -> np.ndarray:
    '''
    derivative of gaussian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum

    Returns:
     first derivative of gaussian edge function
    '''
    tmp = np.exp(-4*np.log(2)*(e/fwhm_G)**2)/np.sqrt(np.pi)

    return 2*np.sqrt(np.log(2))/fwhm_G*tmp


def dderiv_edge_g_aux(e: np.ndarray, fwhm_G: float) -> np.ndarray:
    '''
    2nd derivative of gaussian type edge

    Args:
     e: energy
     fwhm_G: full width at half maximum

    Returns:
     2nd derivative of gaussian edge function
    '''
    tmp = -8*np.log(2)*e*np.exp(-4*np.log(2)*(e/fwhm_G)**2) / \
        (np.sqrt(np.pi)*fwhm_G**2)

    return 2*np.sqrt(np.log(2))/fwhm_G*tmp


def deriv_edge_l_aux(e: np.ndarray, fwhm_L: float) -> np.ndarray:
    '''
    1st derivative of lorenzian type edge

    Args:
     e: energy

    Returns:
     first derivative of lorenzian type function
    '''
    tmp = 1/np.pi/(e**2+fwhm_L**2/4)
    return fwhm_L*tmp/2


def dderiv_edge_l_aux(e: np.ndarray, fwhm_L: float) -> np.ndarray:
    '''
    2nd derivative of lorenzian type edge

    Args:
     e: energy

    Returns:
     second derivative of lorenzian type function
    '''
    tmp = -e/np.pi/(e**2+fwhm_L**2/4)**2
    return fwhm_L*tmp


def deriv_voigt_aux(e: np.ndarray, fwhm_G: float, fwhm_L: float) -> np.ndarray:
    '''
    1st derivative of voigt profile

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:(2\\sqrt{2\\log(2)}\\sigma)
     fwhm_L: full width at half maximum of lorenzian part :math:(2\\gamma)

    Returns:
     first derivative of voigt profile
    '''
    if fwhm_G < 1e-8:
        tmp = fwhm_L/np.pi/(e**2+fwhm_L**2/4)**2
        return -e*tmp

    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    if fwhm_L < 1e-8:
        tmp = np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
        return -e/sigma**2*tmp

    z = (e+complex(0, fwhm_L/2))/(sigma*np.sqrt(2))
    return -(z*wofz(z)).real/(sigma**2*np.sqrt(np.pi))


def dderiv_voigt_aux(e: np.ndarray, fwhm_G: float, fwhm_L: float) -> np.ndarray:
    '''
    2nd derivative of voigt profile

    Args:
     e: energy
     fwhm_G: full width at half maximum of gaussian part :math:(2\\sqrt{2\\log(2)}\\sigma)
     fwhm_L: full width at half maximum of lorenzian part :math:(2\\gamma)

    Returns:
     second derivative of voigt profile
    '''

    if fwhm_G < 1e-8:
        tmp = fwhm_L/np.pi/(e**2+fwhm_L**2/4)**2
        return (4*e**2/(e**2+fwhm_L**2/4)-1)*tmp

    sigma = fwhm_G/(2*np.sqrt(2*np.log(2)))
    if fwhm_L < 1e-8:
        tmp = np.exp(-(e/sigma)**2/2)/(sigma*np.sqrt(2*np.pi))
        return 1/sigma**2*((e/sigma)**2-1)*tmp

    z = (e+complex(0, fwhm_L/2))/(sigma*np.sqrt(2))
    f = wofz(z)/(sigma*np.sqrt(2*np.pi))
    return 1/sigma**4 *\
        ((e**2-fwhm_L**2/4-sigma**2)*f.real - e*fwhm_L*f.imag + fwhm_L/(2*np.pi))


def voigt_thy_aux(e: np.ndarray, thy_peak: np.ndarray,
                  fwhm_G: float, fwhm_L: float,
                  deriv_order: int = 0) -> np.ndarray:
    '''
    Calculates derivative of normalized
    voigt broadened theoretically calculated lineshape spectrum

    Args:
        e: energy
        thy_peak: theoretical calculated peak position and intensity
        fwhm_G: full width at half maximum of gaussian shape
        fwhm_L: full width at half maximum of lorenzian shape
        deriv_order({0, 1, 2}): order of derivative

    Returns:
      derivative of normalized voigt broadened theoritical lineshape spectrum
    '''

    v_matrix = np.empty((e.size, thy_peak.shape[0]))

    if deriv_order == 0:
        for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = voigt(e-thy_peak[i, 0], fwhm_G, fwhm_L)
    elif deriv_order == 1:
        for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = deriv_voigt_aux(e-thy_peak[i, 0], fwhm_G, fwhm_L)
    else:
        for i in range(thy_peak.shape[0]):
            v_matrix[:, i] = dderiv_voigt_aux(e-thy_peak[i, 0], fwhm_G, fwhm_L)

    broadened_theory = v_matrix @ thy_peak[:,
                                           1].reshape((thy_peak.shape[0], 1))

    return broadened_theory.flatten()


def static_spectrum(e: np.ndarray,
                    result: StaticResult, deriv_order: int = 0) -> Callable:
    '''
    Evaluates static spectrum from static spectrum
    fitting result.

    Args:
     e: energy at which evaulate static spectrum
     result: static spectrum fitting result
     deriv_order ({0, 1, 2}): order of derivative. [default: 0]

    Returns:
     Evaluated static spectrum

    Note:
     Baseline feature is removed.
    '''

    tmp = np.zeros(e.size)

    if result['model'] == 'voigt' and result['n_voigt'] > 0:
        e0_voigt = result['x'][:result['n_voigt']]
        fwhm_G_voigt = result['x'][result['n_voigt']:2*result['n_voigt']]
        fwhm_L_voigt = result['x'][2*result['n_voigt']:3*result['n_voigt']]
        V = np.empty((result['n_voigt'], e.size))
        if deriv_order == 0:
            for i in range(result['n_voigt']):
                V[i, :] = voigt(e-e0_voigt[i], fwhm_G_voigt[i],
                                fwhm_L_voigt[i])
        elif deriv_order == 1:
            for i in range(result['n_voigt']):
                V[i, :] = deriv_voigt_aux(e-e0_voigt[i], fwhm_G_voigt[i],
                                          fwhm_L_voigt[i])
        else:
            for i in range(result['n_voigt']):
                V[i, :] = dderiv_voigt_aux(e-e0_voigt[i], fwhm_G_voigt[i],
                                           fwhm_L_voigt[i])
        tmp = (result['c'][:result['n_voigt']]@V).flatten()
    else:
        if result['policy'] == 'shift':
            for i in range(result['n_voigt']):
                peak_tmp = result['thy_peak'][i].copy()
                peak_tmp[:, 0] = peak_tmp[:, 0]+result['x'][2+i]
                tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                                                         result['x'][0], result['x'][1], deriv_order)
        elif result['policy'] == 'scale':
            for i in range(result['n_voigt']):
                peak_tmp = result['thy_peak'][i].copy()
                peak_tmp[:, 0] = peak_tmp[:, 0]*result['x'][2+i]
                tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                                                         result['x'][0], result['x'][1], deriv_order)
        else:
            for i in range(result['n_voigt']):
                peak_tmp = result['thy_peak'][i].copy()
                peak_tmp[:, 0] = result['x'][2+result['n_voigt']+i]*peak_tmp[:, 0] +\
                    result['x'][2+i]
                tmp = tmp + result['c'][i]*voigt_thy_aux(e, peak_tmp,
                                                         result['x'][0], result['x'][1], deriv_order)

    if result['model'] == 'voigt':
        param_edge_start = 3*result['n_voigt']
    else:
        if result['policy'] in ['scale', 'shift']:
            param_edge_start = 2+result['n_voigt']
        else:
            param_edge_start = 2+2*result['n_voigt']

    if result['edge'] == 'g':
        if deriv_order == 0:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    edge_gaussian(e-result['x'][param_edge_start+i],
                                  result['x'][param_edge_start+result['n_edge']+i])
        elif deriv_order == 1:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    deriv_edge_g_aux(e-result['x'][param_edge_start+i],
                                     result['x'][param_edge_start+result['n_edge']+i])
        else:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    dderiv_edge_g_aux(e-result['x'][param_edge_start+i],
                                      result['x'][param_edge_start+result['n_edge']+i])
    elif result['edge'] == 'l':
        if deriv_order == 0:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    edge_lorenzian(e-result['x'][param_edge_start+i],
                                   result['x'][param_edge_start+result['n_edge']+i])
        elif deriv_order == 1:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    deriv_edge_l_aux(e-result['x'][param_edge_start+i],
                                     result['x'][param_edge_start+result['n_edge']+i])
        else:
            for i in range(result['n_edge']):
                tmp = tmp + result['c'][result['n_voigt']+i] *\
                    dderiv_edge_l_aux(e-result['x'][param_edge_start+i],
                                      result['x'][param_edge_start+result['n_edge']+i])
    return tmp



'''
ads:
submodule for driver routine of associated difference spectrum

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Tuple
import numpy as np
from scipy.linalg import svd
from ..mathfun.A_matrix import make_A_matrix_dmp_osc, make_A_matrix_exp
from ..mathfun.rate_eq import compute_signal_irf


def dads(escan_time: np.ndarray, fwhm: float, tau: np.ndarray, base: Optional[bool] = True,
         irf: Optional[str] = 'g', eta: Optional[float] = None,
         intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) \
        -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate decay associated difference spectrum from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      tau: life time for each component
      base: whether or not include baseline [default: True]
      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`

      eta: mixing parameter for pseudo voigt profile
           (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of dataset

    Returns:
     Tuple of calculated decay associated difference spectrum of each component, estimated error,
     and retrieved energy scan intensity from dads and decay components

    Note:
     To calculate decay associated difference spectrum of n component exponential decay, you should measure at least n+1
     energy scan
    '''
    # initialization
    if base:
        c = np.empty((tau.size+1, intensity.shape[0]))
        dof = escan_time.size - (tau.size+1)
    else:
        c = np.empty((tau.size, intensity.shape[0]))
        dof = escan_time.size - (tau.size)

    if eps is None:
        eps = np.ones_like(intensity)

    c_eps = np.empty_like(c)

    A = make_A_matrix_exp(escan_time, fwhm, tau, base, irf, eta)
    data_scaled = intensity/eps

    # evaluates dads
    cond = 1e-2
    for i in range(intensity.shape[0]):
        A_scaled = np.einsum('j,ij->ij', 1/eps[i, :], A)
        U, s, Vh = svd(A_scaled.T, full_matrices=False)
        mask = s > cond*s[0]
        U_turn = U[:, mask]
        s_turn = s[mask]
        Vh_turn = Vh[mask, :]
        c[:, i] = np.einsum('j,ij->ij', 1/s_turn,
                            Vh_turn.T) @ (U_turn.T @ data_scaled[i, :])
        res = data_scaled[i, :] - (c[:, i] @ A_scaled)
        if dof != 0:
            red_chi2 = np.sum(res**2)/dof
            cov_scale = red_chi2 *\
                Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
            c_eps[:, i] = np.sqrt(np.diag(cov_scale))

    return c, c_eps, c.T @ A

def dads_osc(escan_time: np.ndarray, fwhm: float, tau: np.ndarray, 
             tau_osc: np.ndarray, period_osc: np.ndarray, base: Optional[bool] = True,
             irf: Optional[str] = 'g', eta: Optional[float] = None,
             intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) \
             -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate decay associated difference spectrum with damped oscillation 
    from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      tau: life time for each component
      tau_osc: life time for each oscillation component
      period_osc: period of each oscillation
      base: whether or not include baseline [default: True]
      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`

      eta: mixing parameter for pseudo voigt profile
           (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of dataset

    Returns:
     1. decay associated differenece spectrum
     2. damped oscillation associated difference spectrum
     3. estimated error of dads
     4. estimated error of doads
     5. retrieved energy scan from dads
     6. retrieved energy scan from doads

    Note:
     1. To calculate decay associated difference spectrum of n component of exponential decay and 
     m component of damped oscillation, you should measure at least :math:`n+2m+1` energy scan
     2. 0 to m-1 raw of doads is doads of cosine component and m to :math:`2m-1` raw of doads is doads of sine component 
    '''
    # initialization
    if base:
        c = np.empty((tau.size+1+2*tau_osc.size, intensity.shape[0]))
        A = np.empty((tau.size+2*tau_osc.size+1, escan_time.size))
        dof = escan_time.size - (tau.size+1+2*tau_osc.size)
        idx_osc = tau.size+1
    else:
        c = np.empty((tau.size+2*tau_osc.size, intensity.shape[0]))
        A = np.empty((tau.size+2*tau_osc.size, escan_time.size))
        dof = escan_time.size - (tau.size+2*tau_osc.size)
        idx_osc = tau.size

    if eps is None:
        eps = np.ones_like(intensity)

    c_eps = np.empty_like(c)

    A[:idx_osc, :] = make_A_matrix_exp(escan_time, fwhm, tau, base, irf, eta)
    A[idx_osc:, :] = make_A_matrix_dmp_osc(escan_time, fwhm, tau_osc, 
    period_osc, irf, eta)
    data_scaled = intensity/eps

    # evaluates dads
    cond = 1e-2
    for i in range(intensity.shape[0]):
        A_scaled = np.einsum('j,ij->ij', 1/eps[i, :], A)
        U, s, Vh = svd(A_scaled.T, full_matrices=False)
        mask = s > cond*s[0]
        U_turn = U[:, mask]
        s_turn = s[mask]
        Vh_turn = Vh[mask, :]
        c[:, i] = np.einsum('j,ij->ij', 1/s_turn,
                            Vh_turn.T) @ (U_turn.T @ data_scaled[i, :])
        res = data_scaled[i, :] - (c[:, i] @ A_scaled)
        if dof != 0:
            red_chi2 = np.sum(res**2)/dof
            cov_scale = red_chi2 *\
                Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
            c_eps[:, i] = np.sqrt(np.diag(cov_scale))

    return c[:idx_osc,:], c[idx_osc:, :], \
           c_eps[:idx_osc, :], c_eps[idx_osc:, :], \
           c[:idx_osc, :].T @ A[:idx_osc, :], \
           c[idx_osc:, :].T @ A[idx_osc:, :] 


def sads(escan_time: np.ndarray, fwhm: float, eigval: np.ndarray, V: np.ndarray, c: np.ndarray,
         exclude: Optional[str] = None, irf: Optional[str] = 'g', eta: Optional[float] = None,
         intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate species associated difference spectrum from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      eigval: eigenvalue of rate equation matrix
      V: eigenvector of rate equation matrix
      c: coefficient to match initial condition of rate equation
      exclude: exclude either 'first' or 'last' element or both 'first' and 'last' element.

               * 'first' : exclude first element
               * 'last' : exclude last element
               * 'first_and_last' : exclude both first and last element
               * None : Do not exclude any element [default]

      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`

      eta: mixing parameter for pseudo voigt profile (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of data

    Returns:
     Tuple of calculated species associated difference spectrum of each component, estimated error and
     retrieved intensity of energy scan from sads and model excited state components

    Note:
     1. eigval, V, c should be obtained from solve_model
     2. To calculate species associated difference spectrum of n excited state species, you should measure at least n+1 energy scan
     3. Difference spectrum of ground state is zero, so ground state species should be excluded from rate equation or via exclude option.
    '''
    # initialization
    if exclude is None:
        diff_abs = np.empty((eigval.size, intensity.shape[0]))
        dof = escan_time.size - eigval.size
    elif exclude in ['first', 'last']:
        diff_abs = np.empty((eigval.size-1, intensity.shape[0]))
        dof = escan_time.size - (eigval.size-1)
    else:
        diff_abs = np.empty((eigval.size-2, intensity.shape[0]))
        dof = escan_time.size - (eigval.size-2)

    if eps is None:
        eps = np.ones_like(intensity)

    diff_abs_eps = np.empty_like(diff_abs)

    A = compute_signal_irf(escan_time, eigval, V, c, fwhm, irf, eta)
    if exclude == 'first':
        B = A[1:, :]
    elif exclude == 'last':
        B = A[:-1, :]
    elif exclude == 'first_and_last':
        B = A[1:-1, :]
    else:
        B = A

    data_scaled = intensity/eps

    # evaluates sads
    cond = 1e-2
    for i in range(intensity.shape[0]):
        A_scaled = np.einsum('j,ij->ij', 1/eps[i, :], B)
        U, s, Vh = svd(A_scaled.T, full_matrices=False)
        mask = s > cond*s[0]
        U_turn = U[:, mask]
        s_turn = s[mask]
        Vh_turn = Vh[mask, :]
        cov = Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
        diff_abs[:, i] = np.einsum('j,ij->ij', 1/s_turn,
                              Vh_turn.T) @ (U_turn.T @ data_scaled[i, :])
        res = data_scaled[i, :] - (diff_abs[:, i] @ A_scaled)
        if dof != 0:
            red_chi2 = np.sum(res**2)/dof
            diff_abs_eps[:, i] = np.sqrt(red_chi2*np.diag(cov))

    return diff_abs, diff_abs_eps, diff_abs.T @ B

def sads_osc(escan_time: np.ndarray, fwhm: float, 
             eigval: np.ndarray, V: np.ndarray, c: np.ndarray,
             tau_osc: np.ndarray, period_osc: np.ndarray,
             exclude: Optional[str] = None, 
             irf: Optional[str] = 'g', eta: Optional[float] = None,
             intensity: Optional[np.ndarray] = None, eps: Optional[np.ndarray] = None) \
             -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    '''
    Calculate species associated difference spectrum with damped oscillation
    from experimental energy scan data

    Args:
      escan_time: time delay for each energy scan data
      fwhm: full width at half maximum of instrumental response function
      eigval: eigenvalue of rate equation matrix
      V: eigenvector of rate equation matrix
      c: coefficient to match initial condition of rate equation
      tau_osc: life time for each oscillation component
      period_osc: period of each oscillation
      exclude: exclude either 'first' or 'last' element or both 'first' and 'last' element.

               * 'first' : exclude first element
               * 'last' : exclude last element
               * 'first_and_last' : exclude both first and last element
               * None : Do not exclude any element [default]

      irf: shape of instrumental response function [default: g]

           * 'g': normalized gaussian distribution,
           * 'c': normalized cauchy distribution,
           * 'pv': pseudo voigt profile :math:`(1-\\eta)g(t, {fwhm}) + \\eta c(t, {fwhm})`

      eta: mixing parameter for pseudo voigt profile (only needed for pseudo voigt profile)
      intensity: intensity of energy scan dataset
      eps: standard error of data

    Returns:
     1. species associated difference spectrum
     2. damped oscillation asscoiated difference spectrum
     3. estimated error of sads
     4. estimated error of doads
     5. retrieved energy scan from sads
     6. retrieved energy scan from doads

    Note:
     1. eigval, V, c should be obtained from solve_model
     2. To calculate species associated difference spectrum of n excited state species with m damped oscillation component, 
     you should measure at least :math:`n+2m+1` energy scan
     3. Difference spectrum of ground state is zero, 
     so ground state species should be excluded from rate equation or via exclude option.
     4. 0 to m-1 raw of doads is doads of cosine component and m to :math:`2m-1` raw of doads is doads of sine component
    '''
    # initialization
    if exclude is None:
        B = np.empty((eigval.size+2*tau_osc.size, escan_time.size))
        diff_abs = np.empty((eigval.size+2*tau_osc.size, intensity.shape[0]))
        dof = escan_time.size - (eigval.size+2*tau_osc.size)
        idx_osc = eigval.size
    elif exclude in ['first', 'last']:
        B = np.empty((eigval.size-1+2*tau_osc.size, escan_time.size))
        diff_abs = np.empty((eigval.size-1+2*tau_osc.size, intensity.shape[0]))
        dof = escan_time.size - (eigval.size-1+2*tau_osc.size)
        idx_osc = eigval.size-1
    else:
        B = np.empty((eigval.size-2+2*tau_osc.size, escan_time.size))
        diff_abs = np.empty((eigval.size-2, intensity.shape[0]))
        dof = escan_time.size - (eigval.size-2+2*tau_osc.size)
        idx_osc = eigval.size-2

    if eps is None:
        eps = np.ones_like(intensity)

    diff_abs_eps = np.empty_like(diff_abs)

    A = compute_signal_irf(escan_time, eigval, V, c, fwhm, irf, eta)
    if exclude == 'first':
        B[:idx_osc, :] = A[1:, :]
    elif exclude == 'last':
        B[:idx_osc, :] = A[:-1, :]
    elif exclude == 'first_and_last':
        B[:idx_osc, :] = A[1:-1, :]
    else:
        B[:idx_osc, :] = A
    
    B[idx_osc:, :] = make_A_matrix_dmp_osc(escan_time, fwhm, 
    tau_osc, period_osc, irf, eta)

    data_scaled = intensity/eps

    # evaluates sads
    cond = 1e-2
    for i in range(intensity.shape[0]):
        A_scaled = np.einsum('j,ij->ij', 1/eps[i, :], B)
        U, s, Vh = svd(A_scaled.T, full_matrices=False)
        mask = s > cond*s[0]
        U_turn = U[:, mask]
        s_turn = s[mask]
        Vh_turn = Vh[mask, :]
        cov = Vh_turn.T @ np.einsum('i,ij->ij', 1/s_turn**2, Vh_turn)
        diff_abs[:, i] = np.einsum('j,ij->ij', 1/s_turn,
                              Vh_turn.T) @ (U_turn.T @ data_scaled[i, :])
        res = data_scaled[i, :] - (diff_abs[:, i] @ A_scaled)
        if dof != 0:
            red_chi2 = np.sum(res**2)/dof
            diff_abs_eps[:, i] = np.sqrt(red_chi2*np.diag(cov))

    return diff_abs[:idx_osc, :], diff_abs[idx_osc:, :], \
           diff_abs_eps[:idx_osc, :], diff_abs_eps[idx_osc:, :], \
           diff_abs[:idx_osc, :].T @ B[:idx_osc, :], \
           diff_abs[idx_osc:, :].T @ B[idx_osc:, :]
        



'''
res_raise:
submodule for residual function and gradient for fitting time delay scan with the
convolution of sum of raise_model and instrumental response function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from ..mathfun.irf import calc_eta, deriv_eta
from ..mathfun.irf import calc_fwhm, deriv_fwhm
from ..mathfun.A_matrix import make_A_matrix_gau, make_A_matrix_cauchy, fact_anal_A
from ..mathfun.exp_conv_irf import deriv_exp_conv_gau, deriv_exp_conv_cauchy
from ..mathfun.exp_conv_irf import deriv_exp_sum_conv_gau, deriv_exp_sum_conv_cauchy

# residual and gradient function for exponential decay model


def residual_raise(x0: np.ndarray, base: bool, irf: str,
                   t: Optional[Sequence[np.ndarray]] = None,
                   intensity: Optional[Sequence[np.ndarray]] = None, eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_raise
    scipy.optimize.least_squares compatible vector residual function for fitting multiple set of time delay scan with the
    convolution of raise_model :math:`(\\exp(-t/\\tau_{i+1})-\\exp(-t/\\tau_1))` and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        fwhm = calc_fwhm(x0[0], x0[1])
        eta = calc_eta(x0[0], x0[1])

    num_t0 = 0
    count = 0
    for d in intensity:
        num_t0 = d.shape[1] + num_t0
        count = count + d.size

    chi = np.empty(count)
    tau = x0[num_irf+num_t0:]
    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                A = A_gau + eta*(A_cauchy-A_gau)
            A[1:, :] = A[1:, :] - A[0, :]
            c = fact_anal_A(A[1:, :], d[:, j], e[:, j])
            chi[end:end+d.shape[0]] = ((c@A[1:, :]) - d[:, j])/e[:, j]

            end = end + d.shape[0]
            t0_idx = t0_idx + 1

    return chi


def res_grad_raise(x0: np.ndarray, num_comp: int, base: bool, irf: str,
                   fix_param_idx: Optional[np.ndarray] = None,
                   t: Optional[Sequence[np.ndarray]] = None,
                   intensity: Optional[Sequence[np.ndarray]] = None,
                   eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_raise
    scipy.optimize.minimize compatible scalar residual and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of raise_model :math:`(\\exp(-t/\\tau_{i+1})-\\exp(-t/\\tau_1))` and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{scan}`: time zero of each scan
        * :math:`2+N_{scan}` to :math:`2+N_{scan}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{scan}`: time zero of each scan
        * :math:`3+N_{scan}` to :math:`3+N_{scan}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''
    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_fwhm(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])

    num_t0 = 0
    count = 0
    for d in intensity:
        num_t0 = num_t0 + d.shape[1]
        count = count + d.size

    tau = x0[num_irf+num_t0:]

    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    num_param = num_irf+num_t0+num_comp
    chi = np.empty(count)
    df = np.empty((count, tau.size+num_irf))
    grad = np.empty(num_param)

    end = 0
    t0_idx = num_irf

    for ti, d, e in zip(t, intensity, eps):
        step = d.shape[0]
        for j in range(d.shape[1]):
            t0 = x0[t0_idx]
            if irf == 'g':
                A = make_A_matrix_gau(ti-t0, fwhm, k)
            elif irf == 'c':
                A = make_A_matrix_cauchy(ti-t0, fwhm, k)
            else:
                A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
                A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
                diff = A_cauchy-A_gau
                A = A_gau + eta*diff
            A[1:, :] = A[1:, :] - A[0, :]
            c = fact_anal_A(A[1:, :], d[:, j], e[:, j])
            chi[end:end+step] = (c@A[1:, :]-d[:, j])/e[:, j]

            c_grad = np.hstack((np.array([-np.sum(c)]), c))

            if irf == 'g':
                grad_tmp = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c_grad, base)
            elif irf == 'c':
                grad_tmp = deriv_exp_sum_conv_cauchy(
                    ti-t0, fwhm, 1/tau, c, base)
            else:
                grad_gau = deriv_exp_sum_conv_gau(ti-t0, fwhm, 1/tau, c_grad, base)
                grad_cauchy = deriv_exp_sum_conv_cauchy(
                    ti-t0, fwhm, 1/tau, c_grad, base)
                grad_tmp = grad_gau + eta*(grad_cauchy-grad_gau)

            grad_tmp = np.einsum('i,ij->ij', 1/e[:, j], grad_tmp)
            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_tmp[:, 1]
            else:
                cdiff = (c_grad@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_tmp[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_tmp[:, 1]+deta_L*cdiff
            grad[t0_idx] = -chi[end:end+step]@grad_tmp[:, 0]
            df[end:end+step,
                num_irf:] = np.einsum('j,ij->ij', -1/tau**2, grad_tmp[:, 2:])

            end = end + step
            t0_idx = t0_idx + 1

    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_t0] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0

    return np.sum(chi**2)/2, grad

def residual_raise_same_t0(x0: np.ndarray, base: bool, irf: str,
                           t: Optional[Sequence[np.ndarray]] = None,
                           intensity: Optional[Sequence[np.ndarray]] = None,
                           eps: Optional[Sequence[np.ndarray]] = None) -> np.ndarray:
    '''
    residual_raise_same_t0
    scipy.optimize.least_squares compatible vector residual function
    for fitting multiple set of time delay scan with the
    sum of convolution of raise_model :math:`(\\exp(-t/\\tau_{i+1})-\\exp(-t/\\tau_1))`
    and instrumental response function
    Set Time Zero of every time dset in same dataset same

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{dset}`: time zero of each data set
        * :math:`2+N_{dset}` to :math:`2+N_{dset}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{dset}`: time zero of each data set
        * :math:`3+N_{dset}` to :math:`3+N_{dset}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Residual vector
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
    else:
        num_irf = 2
        fwhm = calc_fwhm(x0[0], x0[1])
        eta = calc_eta(x0[0], x0[1])

    num_dataset = len(t)
    count = 0
    for i in range(num_dataset):
        count = count + intensity[i].size

    chi = np.empty(count)
    tau = x0[num_irf+num_dataset:]
    if not base:
        k = 1/tau
    else:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        t0 = x0[t0_idx]
        if irf == 'g':
            A = make_A_matrix_gau(ti-t0, fwhm, k)
        elif irf == 'c':
            A = make_A_matrix_cauchy(ti-t0, fwhm, k)
        else:
            A_gau = make_A_matrix_gau(ti-t0, fwhm, k)
            A_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
            A = A_gau + eta*(A_cauchy-A_gau)
        A[1:, :] = A[1:, :] - A[0, :]
        for j in range(d.shape[1]):
            c = fact_anal_A(A[1:, :], d[:, j], e[:, j])
            chi[end:end+d.shape[0]] = ((c@A[1:, :]) - d[:, j])/e[:, j]

            end = end + d.shape[0]
        t0_idx = t0_idx + 1

    return chi

def res_grad_raise_same_t0(x0: np.ndarray, num_comp: int, base: bool, irf: str,
                           fix_param_idx: Optional[np.ndarray] = None,
                           t: Optional[Sequence[np.ndarray]] = None,
                           intensity: Optional[Sequence[np.ndarray]] = None,
                           eps: Optional[Sequence[np.ndarray]] = None) -> Tuple[np.ndarray, np.ndarray]:
    '''
    res_grad_raise_same_t0
    scipy.optimize.minimize compatible scalar residual
    and its gradient function for fitting multiple set of time delay scan with the
    sum of convolution of raise_model :math:`(\\exp(-t/\\tau_{i+1})-\\exp(-t/\\tau_1))`
    and instrumental response function

    Args:
     x0: initial parameter,
      if irf == 'g','c':

        * 1st: fwhm_(G/L)
        * 2nd to :math:`2+N_{dset}`: time zero of each dataset
        * :math:`2+N_{dset}` to :math:`2+N_{dset}+N_{\\tau}`: time constant of each decay component

      if irf == 'pv':

        * 1st and 2nd: fwhm_G, fwhm_L
        * 3rd to :math:`3+N_{dset}`: time zero of each dataset
        * :math:`3+N_{dset}` to :math:`3+N_{dset}+N_{\\tau}`: time constant of each decay component

     num_comp: number of exponential decay component (except base)
     base: whether or not include baseline (i.e. very long lifetime component)
     irf: shape of instrumental response function

          * 'g': normalized gaussian distribution,
          * 'c': normalized cauchy distribution,
          * 'pv': pseudo voigt profile :math:`(1-\\eta)g(f) + \\eta c(f)`

        For pseudo voigt profile, the mixing parameter :math:`\\eta(f_G, f_L)` and
        uniform fwhm paramter :math:`f(f_G, f_L)` are calculated by `calc_eta` and `calc_fwhm` routine
     t: time points for each data set
     fix_param_idx: index for fixed parameter (masked array for `x0`)
     intensity: sequence of intensity of datasets
     eps: sequence of estimated error of datasets

    Returns:
     Tuple of scalar residual function :math:`(\\frac{1}{2}\\sum_i {res}^2_i)` and its gradient
    '''

    x0 = np.atleast_1d(x0)

    if irf in ['g', 'c']:
        num_irf = 1
        fwhm = x0[0]
        eta = None
    else:
        num_irf = 2
        eta = calc_eta(x0[0], x0[1])
        fwhm = calc_fwhm(x0[0], x0[1])
        deta_G, deta_L = deriv_eta(x0[0], x0[1])
        dfwhm_G, dfwhm_L = deriv_fwhm(x0[0], x0[1])

    num_dataset = len(t)
    count = 0
    for i in range(num_dataset):
        count = count + intensity[i].size

    tau = x0[num_irf+num_dataset:num_irf+num_dataset+num_comp]

    if base:
        k = np.empty(tau.size+1)
        k[:-1] = 1/tau
        k[-1] = 0
    else:
        k = 1/tau

    num_param = num_irf+num_dataset+num_comp

    chi = np.empty(count)
    df = np.zeros((count, num_irf+num_comp))
    grad = np.zeros(num_param)

    end = 0
    t0_idx = num_irf
    for ti, d, e in zip(t, intensity, eps):
        step = d.shape[0]
        A = np.empty((num_comp+1*base, step))
        A_grad_decay = np.empty((num_comp+1*base, step, 3))
        grad_decay = np.empty((step, num_comp+2))

        t0 = x0[t0_idx]
        if irf == 'g':
            A[:num_comp+1*base, :] = make_A_matrix_gau(ti-t0, fwhm, k)
            for i in range(num_comp):
                A_grad_decay[i, :, :] = deriv_exp_conv_gau(ti-t0, fwhm, k[i])
            if base:
                A_grad_decay[-1, :, :] = deriv_exp_conv_gau(ti-t0, fwhm, 0)
        elif irf == 'c':
            A[:num_comp+1*base, :] = make_A_matrix_cauchy(ti-t0, fwhm, k)
            for i in range(num_comp):
                A_grad_decay[i, :, :] = deriv_exp_conv_cauchy(ti-t0, fwhm, k[i])
            if base:
                A_grad_decay[-1, :, :] = deriv_exp_conv_cauchy(ti-t0, fwhm, 0)
        else:
            tmp_gau = make_A_matrix_gau(ti-t0, fwhm, k)
            tmp_cauchy = make_A_matrix_cauchy(ti-t0, fwhm, k)
            diff = tmp_cauchy-tmp_gau
            A[:num_comp+1*base, :] = tmp_gau + eta*diff
            for i in range(num_comp):
                tmp_grad_gau = deriv_exp_conv_gau(ti-t0, fwhm, k[i])
                tmp_grad_cauchy = deriv_exp_conv_cauchy(ti-t0, fwhm, k[i])
                A_grad_decay[i, :, :] = tmp_grad_gau + eta*(tmp_grad_cauchy-tmp_grad_gau)
            if base:
                tmp_grad_gau = deriv_exp_conv_gau(ti-t0, fwhm, 0)
                tmp_grad_cauchy = deriv_exp_conv_cauchy(ti-t0, fwhm, 0)
                A_grad_decay[-1, :, :] = tmp_grad_gau + eta*(tmp_grad_cauchy-tmp_grad_gau)
 
        A[1:, :] = A[1:, :] - A[0, :]

        for j in range(d.shape[1]):
            c = fact_anal_A(A[1:, :], d[:, j], e[:, j])
            chi[end:end+step] = (c@A[1:, :]-d[:, j])/e[:, j]

            c_grad = np.hstack((np.array([-np.sum(c)]), c))

            grad_decay[:, :2] = \
                np.tensordot(c_grad[:num_comp+1*base], A_grad_decay[:, :, :2], axes=1)
            for i in range(num_comp):
                grad_decay[:, 2+i] = c_grad[i]*A_grad_decay[i, :, 2]

            grad_decay = np.einsum('i,ij->ij', 1/e[:, j], grad_decay)

            if irf in ['g', 'c']:
                df[end:end+step, 0] = grad_decay[:, 1]

            else:
                cdiff = (c_grad[:num_comp+1*base]@diff)/e[:, j]
                df[end:end+step, 0] = dfwhm_G*grad_decay[:, 1]+deta_G*cdiff
                df[end:end+step, 1] = dfwhm_L*grad_decay[:, 1]+deta_L*cdiff

            grad[t0_idx] = grad[t0_idx] -chi[end:end+step]@grad_decay[:, 0]
            df[end:end+step, num_irf:num_irf+num_comp] = \
                np.einsum('j,ij->ij', -1/tau**2, grad_decay[:, 2:])

            end = end + step

        t0_idx = t0_idx + 1

    mask = np.ones(num_param, dtype=bool)
    mask[num_irf:num_irf+num_dataset] = False
    grad[mask] = chi@df

    if fix_param_idx is not None:
        grad[fix_param_idx] = 0

    return np.sum(chi**2)/2, grad



'''
TRXASprefitpack:
package for TRXAS pre- and fitting process which aims for the first order dynamics
TRXAS stands for Time Resolved X-ray Absorption Spectroscopy

:copyright: 2021-2022 by pistack (Junho Lee)
:license: LGPL3.
'''

from .mathfun import *
from .driver import *
from .res import *

__version__ = '0.7.2'



'''
_static_voigt:
submodule for static spectrum with the
sum of voigt function, edge function and baseline function

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

from typing import Optional, Sequence, Tuple
import numpy as np
from numpy.polynomial.legendre import legval
from .static_result import StaticResult
from ._ampgo import ampgo
from scipy.optimize import basinhopping
from scipy.optimize import least_squares
from ..mathfun.peak_shape import voigt, edge_gaussian, edge_lorenzian
from ..mathfun.A_matrix import fact_anal_A
from ..res.parm_bound import set_bound_e0, set_bound_t0
from ..res.res_voigt import residual_voigt, res_grad_voigt

GLBSOLVER = {'basinhopping': basinhopping, 'ampgo': ampgo}


def fit_static_voigt(e0_init: np.ndarray, fwhm_G_init: np.ndarray, fwhm_L_init: np.ndarray,
                     edge: Optional[str] = None,
                     edge_pos_init: Optional[np.ndarray] = None,
                     edge_fwhm_init: Optional[np.ndarray] = None,
                     base_order: Optional[int] = None,
                     method_glb: Optional[str] = None,
                     method_lsq: Optional[str] = 'trf',
                     kwargs_glb: Optional[dict] = None,
                     kwargs_lsq: Optional[dict] = None,
                     bound_e0: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_fwhm_G: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_fwhm_L: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_edge_pos: Optional[Sequence[Tuple[float, float]]] = None,
                     bound_edge_fwhm: Optional[Sequence[Tuple[float, float]]] = None,
                     e: Optional[np.ndarray] = None,
                     intensity: Optional[np.ndarray] = None,
                     eps: Optional[np.ndarray] = None) -> StaticResult:
    '''
    driver routine for fitting static spectrum with sum of voigt component, edge and
    polynomial baseline.

    It separates linear and non-linear part of the optimization problem to solve non linear least sequare
    optimization problem efficiently.

    Moreover this driver uses two step method to search best parameter, its covariance and
    estimated parameter error.

    Step 1. (method_glb)
    Use global optimization to find rough global minimum of our objective function.
    In this stage, it use analytic gradient for scalar residual function.

    Step 2. (method_lsq)
    Use least squares optimization algorithm to refine global minimum of objective function and approximate covariance matrix.
    Because of linear and non-linear seperation scheme, the analytic jacobian for vector residual function is hard to optain.
    Thus, in this stage, it uses numerical jacobian.

    Args:
     e0_init (np.ndarray): initial peak position of each voigt component
     fwhm_G_init (np.ndarray): initial gaussian part of fwhm parameter of each voigt component
     fwhm_L_init (np.ndarray): initial lorenzian part of fwhm parameter of each voigt component
     edge ({'g', 'l'}): type of edge function. If edge is not set, edge feature is not included.
     edge_pos_init (np.ndarray): initial edge position
     edge_fwhm_init (np.ndarray): initial fwhm parameter of edge
     method_glb ({None, 'basinhopping', 'ampgo'}): Method for global optimization Algorithm.
     method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                           minimization (refinement of global optimization solution)
     kwargs_glb: keyward arguments for global optimization solver
     kwargs_lsq: keyward arguments for least square optimization solver
     bound_e0 (sequence of tuple): boundary for each voigt componet. If upper and lower bound are same,
      driver assumes that the parameter is fixed during optimization. If `bound_e0` is `None`,
      the upper and lower bound are given by `set_bound_e0`.
     bound_fwhm_G (sequence of tuple): boundary for fwhm_G parameter.
      If `bound_fwhm_G` is `None`, the upper and lower bound are given as `(fwhm_G/2, 2*fwhm_G)`.
     bound_fwhm_L (sequence of tuple): boundary for fwhm_L parameter.
      If `bound_fwhm_L` is `None`, the upper and lower bound are given as `(fwhm_L/2, 2*fwhm_L)`.
     bound_edge_pos (sequence of tuple): boundary for edge position,
      if `bound_edge_pos` is `None` and `edge` is set, the upper and lower bound are given by `set_bound_t0`.
     bound_edge_fwhm (sequence of tuple): boundary for fwhm parameter of edge feature.
      If `bound_edge_fwhm` is `None`, the upper and lower bound are given as `(edge_fwhm/2, 2*edge_fwhm)`.
     e (np.narray): energy range for data
     intensity (np.ndarray): intensity of static spectrum data
     eps (np.ndarray): estimated errors of static spectrum data

     Returns:
      StaticResult class object
     Note:

      * if initial fwhm_G is zero then such voigt component is treated as lorenzian component
      * if initial fwhm_L is zero then such voigt component is treated as gaussian component
    '''

    if method_glb is not None and method_glb not in ['basinhopping', 'ampgo']:
        raise Exception('Unsupported global optimization Method, Supported global optimization Methods are ampgo and basinhopping')
    if method_lsq not in ['trf', 'lm', 'dogbox']:
        raise Exception('Invalid local least square minimizer solver. It should be one of [trf, lm, dogbox]')
    if edge is not None and edge not in  ['g', 'l']:
        raise Exception('Invalid Edge type.')

    if e0_init is None:
        num_voigt = 0
        num_param = 0
    else:
        num_voigt = e0_init.size
        num_param = 3*num_voigt

    num_comp = num_voigt
    num_edge = 0
    if edge is not None:
        num_edge = edge_pos_init.size
        num_comp = num_comp+num_edge
        num_param = num_param+2*num_edge

    if base_order is not None:
        num_comp = num_comp + base_order + 1

    param = np.empty(num_param, dtype=float)
    fix_param_idx = np.empty(num_param, dtype=bool)

    param[:num_voigt] = e0_init
    param[num_voigt:2*num_voigt] = fwhm_G_init
    param[2*num_voigt:3*num_voigt] = fwhm_L_init
    if edge is not None:
        param[3*num_voigt:3*num_voigt+num_edge] = edge_pos_init
        param[3*num_voigt+num_edge:] = edge_fwhm_init

    bound = num_param*[None]

    if bound_e0 is None:
        for i in range(num_voigt):
            bound[i] = set_bound_e0(e0_init[i], fwhm_G_init[i], fwhm_L_init[i])
    else:
        bound[:num_voigt] = bound_e0

    if bound_fwhm_G is None:
        for i in range(num_voigt):
            bound[i+num_voigt] = (fwhm_G_init[i]/2, 2*fwhm_G_init[i])
    else:
        bound[num_voigt: 2*num_voigt] = bound_fwhm_G

    if bound_fwhm_L is None:
        for i in range(num_voigt):
            bound[i+2*num_voigt] = (fwhm_L_init[i]/2, 2*fwhm_L_init[i])
    else:
        bound[2*num_voigt:3*num_voigt] = bound_fwhm_L

    if edge is not None:
        if bound_edge_pos is None:
            for i in range(num_edge):
                bound[3*num_voigt+i] = \
                    set_bound_t0(edge_pos_init[i], edge_fwhm_init[i])
        else:
            bound[3*num_voigt:3*num_voigt+num_edge] = bound_edge_pos
        if bound_edge_fwhm is None:
            for i in range(num_edge):
                bound[3*num_voigt+num_edge+i] = \
                    (edge_fwhm_init[i]/2, 2*edge_fwhm_init[i])
        else:
            bound[3*num_voigt+num_edge:] = bound_edge_fwhm

    for i in range(num_param):
        fix_param_idx[i] = (bound[i][0] == bound[i][1])

    if method_glb is not None:
        go_args = (num_voigt, edge, num_edge, base_order, fix_param_idx,
                   e, intensity, eps)
        min_go_kwargs = {'args': go_args, 'jac': True, 'bounds': bound}
        if kwargs_glb is not None:
            minimizer_kwargs = kwargs_glb.pop('minimizer_kwargs', None)
            if minimizer_kwargs is None:
                kwargs_glb['minimizer_kwargs'] = min_go_kwargs
            else:
                minimizer_kwargs['args'] = go_args
                minimizer_kwargs['jac'] = True
                minimizer_kwargs['bouns'] = bound
                kwargs_glb['minimizer_kwargs'] = minimizer_kwargs
        else:
            kwargs_glb = {'minimizer_kwargs': min_go_kwargs}
        res_go = GLBSOLVER[method_glb](res_grad_voigt, param, **kwargs_glb)
    else:
        res_go = {}
        res_go['x'] = param
        res_go['message'] = None
        res_go['nfev'] = 0

    param_gopt = res_go['x']

    lsq_args = (num_voigt, edge, num_edge, base_order, e, intensity, eps)

    if kwargs_lsq is not None:
        _ = kwargs_lsq.pop('args', None)
        _ = kwargs_lsq.pop('kwargs', None)
        kwargs_lsq['args'] = lsq_args
    else:
        kwargs_lsq = {'args': lsq_args}

    bound_tuple = (num_param*[None], num_param*[None])
    for i in range(num_param):
        bound_tuple[0][i] = bound[i][0]
        bound_tuple[1][i] = bound[i][1]
        if bound[i][0] == bound[i][1]:
            if bound[i][0] > 0:
                bound_tuple[1][i] = bound[i][1]*(1+1e-8)+1e-16
            else:
                bound_tuple[1][i] = bound[i][1]*(1-1e-8)+1e-16

    # jacobian of vector residual function is inaccurate
    res_lsq = least_squares(residual_voigt, param_gopt,
                            method=method_lsq, bounds=bound_tuple, **kwargs_lsq)
    param_opt = res_lsq['x']

    e0_opt = param_opt[:num_voigt]
    fwhm_G_opt = param_opt[num_voigt:2*num_voigt]
    fwhm_L_opt = param_opt[2*num_voigt:3*num_voigt]
    e0_edge_opt = param_opt[3*num_voigt:3*num_voigt+num_edge]
    fwhm_edge_opt = param_opt[3*num_voigt+num_edge:]

  # Calc individual chi2
    chi = res_lsq['fun']
    num_param_tot = num_comp+num_param-np.sum(fix_param_idx)
    chi2 = 2*res_lsq['cost']
    red_chi2 = chi2/(chi.size-num_param_tot)

    param_name = np.empty(param_opt.size, dtype=object)
    for i in range(num_voigt):
        param_name[i] = f'e0_{i+1}'
        param_name[num_voigt+i] = f'fwhm_(G, {i+1})'
        param_name[2*num_voigt+i] = f'fwhm_(L, {i+1})'

    if edge is not None:
        for i in range(num_edge):
            param_name[3*num_voigt+i] = f'E0_({edge}, {i+1})'
        if edge == 'g':
            for i in range(num_edge):
                param_name[3*num_voigt+num_edge+i] = \
                    f'fwhm_(G, edge, {i+1})'
        elif edge == 'l':
            for i in range(num_edge):
                param_name[3*num_voigt+num_edge+i] = \
                    f'fwhm_(L, edge, {i+1})'

    A = np.empty((num_comp, e.size))

    for i in range(num_voigt):
        A[i, :] = voigt(e-e0_opt[i], fwhm_G_opt[i], fwhm_L_opt[i])

    base_start = num_voigt

    if edge is not None:
        base_start = base_start+num_edge
        if edge == 'g':
            for i in range(num_edge):
                A[num_voigt+i, :] = \
                    edge_gaussian(e-e0_edge_opt[i], fwhm_edge_opt[i])
        elif edge == 'l':
            for i in range(num_edge):
                A[num_voigt+i, :] = \
                    edge_lorenzian(e-e0_edge_opt[i], fwhm_edge_opt[i])

    if base_order is not None:
        e_max = np.max(e)
        e_min = np.min(e)
        e_norm = 2*(e-(e_max+e_min)/2)/(e_max-e_min)
        tmp = np.eye(base_order+1)
        A[base_start:, :] = legval(e_norm, tmp, tensor=True)

    c = fact_anal_A(A, intensity, eps)

    fit = c@A

    if edge is not None:
        fit_comp = np.einsum('i,ij->ij', c[:num_voigt+num_edge],
                             A[:num_voigt+num_edge, :])
    else:
        fit_comp = np.einsum('i,ij->ij', c[:num_voigt],
                             A[:num_voigt, :])

    base = None

    if base_order is not None:
        base = c[base_start:]@A[base_start:, :]

    res = intensity - fit

    jac = res_lsq['jac']
    hes = jac.T @ jac
    cov = np.zeros_like(hes)
    n_free_param = np.sum(~fix_param_idx)
    mask_2d = np.einsum('i,j->ij', ~fix_param_idx, ~fix_param_idx)
    cov[mask_2d] = np.linalg.inv(hes[mask_2d].reshape(
        (n_free_param, n_free_param))).flatten()
    cov_scaled = red_chi2*cov
    param_eps = np.sqrt(np.diag(cov_scaled))
    corr = cov_scaled.copy()
    weight = np.einsum('i,j->ij', param_eps, param_eps)
    corr[mask_2d] = corr[mask_2d]/weight[mask_2d]

    result = StaticResult()
    result['model'] = 'voigt'
    result['e'] = e
    result['intensity'] = intensity
    result['eps'] = eps
    result['fit'] = fit
    result['fit_comp'] = fit_comp
    result['res'] = res
    result['base_order'] = base_order
    result['edge'] = edge
    result['n_voigt'] = num_voigt
    result['n_edge'] = num_edge
    result['param_name'] = param_name
    result['x'] = param_opt
    result['bounds'] = bound
    result['base'] = base
    result['c'] = c
    result['chi2'] = chi2
    result['aic'] = chi.size*np.log(chi2/chi.size)+2*num_param_tot
    result['bic'] = chi.size * \
        np.log(chi2/chi.size)+num_param_tot*np.log(chi.size)
    result['red_chi2'] = red_chi2
    result['nfev'] = res_go['nfev'] + res_lsq['nfev']
    result['n_param'] = num_param_tot
    result['num_pts'] = chi.size
    result['jac'] = jac
    result['cov'] = cov
    result['corr'] = corr
    result['cov_scaled'] = cov_scaled
    result['x_eps'] = param_eps
    result['method_lsq'] = method_lsq
    result['message_lsq'] = res_lsq['message']
    result['success_lsq'] = res_lsq['success']

    if result['success_lsq']:
        result['status'] = 0
    else:
        result['status'] = -1

    if method_glb is not None:
        result['method_glb'] = method_glb
        result['message_glb'] = res_go['message'][0]
    else:
        result['method_glb'] = None
        result['message_glb'] = None

    return result



'''
transient_result:
submodule for reporting fitting process of time scan result

:copyright: 2021-2022 by pistack (Junho Lee).
:license: LGPL3.
'''

import h5py as h5
import numpy as np


class TransientResult(dict):
    '''
    Represent results for fitting driver routine

    Attributes:
     model ({'decay', 'dmp_osc', 'both'}): model used for fitting

            `decay`: sum of the convolution of exponential decay and instrumental response function

            `dmp_osc`: sum of the convolution of damped oscillation and instrumental response function

            `both`: sum of `decay` and `dmp_osc` model
     same_t0 (bool): whether or not time zero set to same for every time delay scan in same dataset
     name_of_dset (sequence of str): name of each dataset
     t (sequence of np.ndarray): time range for each dataset
     intensity (sequence of np.ndarray): sequence of datasets of intensity of time delay scan
     eps (sequence of np.ndarray): sequence of datasets for estimated error of time delay scan
     fit (sequence of np.ndarray): fitting curve for each data set
     fit_decay (sequence of np.ndarray): decay part of fitting curve for each data set [model = 'both']
     fit_osc (sequence of np.ndarray): oscillation part of fitting curve for each data set [model = 'both']
     res (sequence of np.ndarray): residual curve (data-fit) for each data set
     irf ({'g', 'c', 'pv'}): shape of instrument response function

          'g': gaussian instrumental response function

          'c': cauchy (lorenzian) instrumental response function

          'pv': pseudo voigt instrumental response function (linear combination of gaussian and lorenzian function)
     fwhm (float): unifrom fwhm parameter for pseudo voigt function :math:`((1-\\eta)*g(t, {fwhm})+\\eta*c{t, {fwhm}})`
     eta (float): mixing parameter for pseudo voigt function :math:`((1-\\eta)*g(t, {fwhm})+\\eta*c{t, {fwhm}})`
     param_name (np.ndarray): name of parameter
     n_decay (int): number of decay components (except baseline feature)
     n_osc (int): number of damped oscillation components
     x (np.ndarray): best parameter
     bounds (sequence of tuple): boundary of each parameter
     base (bool): whether or not use baseline feature in fitting process
     c (sequence of np.ndarray): best weight of each component of each datasets
     phase (sequence of np.ndarray): phase factor of each oscillation component of each datasets [mode = 'dmp_osc', 'both']
     chi2 (float): total chi squared value of fitting
     aic (float): Akaike Information Criterion statistic: :math:`N\\log(\\chi^2/N)+2N_{parm}`
     bic (float): Bayesian Information Criterion statistic: :math:`N\\log(\\chi^2/N)+N_{parm}\\log(N)`
     chi2_ind (np.ndarray): chi squared value of individual time delay scan
     red_chi2 (float): total reduced chi squared value of fitting
     red_chi2_ind (np.ndarray): reduced chi squared value of individul time delay scan
     nfev (int): total number of function evaluation
     n_param (int): total number of effective parameter
     n_param_ind (int): number of parameter which affects fitting quality of indiviual time delay scan
     num_pts (int): total number of data points
     jac (np.ndarray): jacobian of objective function at optimal point
     cov (np.ndarray): covariance matrix (i.e. inverse of :math:`J^T J`)
     cov_scaled (np.ndarray): scaled covariance matrix (i.e. :math:`\\chi^2_{red} \\cdot {cov}`)
     corr (np.ndarray): parameter correlation matrix
     x_eps (np.ndarray): estimated error of parameter
      (i.e. square root of diagonal element of `conv_scaled`)
     method_glb ({'ampgo', 'basinhopping'}):
      method of global optimization used in fitting process
     message_glb (str): messages from global optimization process
     method_lsq ({'trf', 'dogbox', 'lm'}): method of local optimization for least_squares
                                           minimization (refinement of global optimization solution)
     success_lsq (bool): whether or not local least square optimization is successed
     message_lsq (str): messages from local least square optimization process
     status ({0, -1}): status of optimization process

                   `0` : normal termination

                   `-1` : least square optimization process is failed
    '''

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError as e:
            raise AttributeError(name) from e

    __setattr__ = dict.__setitem__
    __delattr__ = dict.__delitem__

    def __dir__(self):
        return list(self.keys())

    def __str__(self, corr_tol: float = 1e-1):
        '''
        Report TransientResult class

        Args:
        result: TransientResult class instance which has fitting result
        corr_tol: parameter correlation greather `corr_tol` would be reported

        Returns:
         string which reports fitting results
        '''

        doc_lst = []
        doc_lst.append('[Model information]')
        doc_lst.append(f"    model : {self['model']}")
        doc_lst.append(f"    irf: {self['irf']}")
        doc_lst.append(f"    fwhm: {self['fwhm']: .4f}")
        doc_lst.append(f"    eta: {self['eta']: .4f}")
        doc_lst.append(f"    base: {self['base']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Method]')
        if self['method_glb'] is not None:
            doc_lst.append(f"    global: {self['method_glb']}")
        doc_lst.append(f"    leastsq: {self['method_lsq']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Status]')
        doc_lst.append(f"    nfev: {self['nfev']}")
        doc_lst.append(f"    status: {self['status']}")
        if self['method_glb'] is not None:
            doc_lst.append(f"    global_opt msg: {self['message_glb']}")
        doc_lst.append(f"    leastsq_opt msg: {self['message_lsq']}")
        doc_lst.append(' ')
        doc_lst.append('[Optimization Results]')
        doc_lst.append(f"    Total Data points: {self['num_pts']}")
        doc_lst.append(
            f"    Number of effective parameters: {self['n_param']}")
        doc_lst.append(
            f"    Degree of Freedom: {self['num_pts']-self['n_param']}")
        doc_lst.append(
            f"    Chi squared: {self['chi2']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    Reduced chi squared: {self['red_chi2']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    AIC (Akaike Information Criterion statistic): {self['aic']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(
            f"    BIC (Bayesian Information Criterion statistic): {self['bic']: .4f}".rstrip('0').rstrip('.'))
        doc_lst.append(' ')
        doc_lst.append('[Parameters]')
        for pn, pv, p_err in zip(self['param_name'], self['x'], self['x_eps']):
            doc_lst.append(
                f'    {pn}: {pv: .8f} +/- {p_err: .8f} ({100*np.abs(p_err/pv): .2f}%)'.rstrip('0').rstrip('.'))
        doc_lst.append(' ')
        doc_lst.append('[Parameter Bound]')
        for pn, pv, bd in zip(self['param_name'], self['x'], self['bounds']):
            doc_lst.append(f'    {pn}: {bd[0]: .8f}'.rstrip('0').rstrip(
                '.')+f' <= {pv: .8f} <= {bd[1]: .8f}'.rstrip('0').rstrip('.'))
        doc_lst.append(' ')

        if self['model'] in ['dmp_osc', 'both']:
            doc_lst.append('[Phase Factor]')
            for i in range(len(self['phase'])):
                doc_lst.append(f"    DataSet {self['name_of_dset'][i]}:")
                row = ['     #tscan']
                for j in range(self['c'][i].shape[1]):
                    row.append(f'tscan_{j+1}')
                doc_lst.append('\t'.join(row))
            for o in range(self['n_osc']):
                row = [f'     dmp_osc {o+1}']
                for l in range(self['phase'][i].shape[1]):
                    row.append(f"{self['phase'][i][o, l]/np.pi: .4f} π")
                doc_lst.append('\t'.join(row))
        doc_lst.append(' ')

        doc_lst.append('[Component Contribution]')
        if self['model'] == 'raise':
            tot_decay = self['n_decay']-1
        else:
            tot_decay = self['n_decay']

        for i in range(len(self['c'])):
            doc_lst.append(f"    DataSet {self['name_of_dset'][i]}:")
            row = ['     #tscan']
            coeff_abs = np.abs(self['c'][i])
            coeff_sum = np.sum(coeff_abs, axis=0)
            coeff_contrib = np.einsum('j,ij->ij', 100/coeff_sum, self['c'][i])
            for j in range(coeff_contrib.shape[1]):
                row.append(f'tscan_{j+1}')
            doc_lst.append('\t'.join(row))
            for d in range(tot_decay):
                row = [f'     decay {d+1}']
                for l in range(coeff_contrib.shape[1]):
                    row.append(f'{coeff_contrib[d, l]: .2f}%')
                doc_lst.append('\t'.join(row))
            
            if self['base']:
                tmp = tot_decay+1
                row = ['     base']
                for l in range(coeff_contrib.shape[1]):
                    row.append(f'{coeff_contrib[tmp-1,l]: .2f}%')
                doc_lst.append('\t'.join(row))

            for o in range(tot_decay, tot_decay+self['n_osc']):
                row = [f'    dmp_osc {o+1-tot_decay}']
                for l in range(coeff_contrib.shape[1]):
                    row.append(f'{coeff_contrib[o, l]: .2f}%')
                doc_lst.append('\t'.join(row))
        doc_lst.append(' ')
        doc_lst.append('[Parameter Correlation]')
        doc_lst.append(f'    Parameter Correlations > {corr_tol: .3f}'.rstrip(
            '0').rstrip('.') + ' are reported.')

        A = np.empty((len(self['x']), len(self['x'])), dtype=object)
        for i in range(len(self['x'])):
            for j in range(len(self['x'])):
                A[i, j] = (i, j)

        mask = (np.abs(self['corr']) > corr_tol)

        for pair in A[mask]:
            if pair[0] > pair[1]:
                tmp_str_lst = [f"    ({self['param_name'][pair[0]]},"]
                tmp_str_lst.append(f"{self['param_name'][pair[1]]})")
                tmp_str_lst.append('=')
                tmp_str_lst.append(
                    f"{self['corr'][pair]: .3f}".rstrip('0').rstrip('.'))
                doc_lst.append(' '.join(tmp_str_lst))

        return '\n'.join(doc_lst)


def save_TransientResult(result: TransientResult, filename: str):
    '''
    save transient fitting result to the h5 file

    Args:
     result: transient fitting result
     filename: filename to store result. It will store result to filename.h5

    Returns:
     h5 file which stores result
    '''

    model_key_lst = ['model', 'irf', 'fwhm', 'eta', 'n_decay', 'n_osc',
                     'base', 'chi2', 'n_param', 'n_param_ind', 'num_pts', 'red_chi2',
                     'aic', 'bic', 'nfev',
                     'method_lsq', 'success_lsq', 'message_lsq', 'status']

    exp_key_lst = ['t', 'intensity', 'eps']
    fit_key_lst = ['fit', 'res', 'c', 'chi2_ind', 'red_chi2_ind']
    exp_dir_lst = ['time_delay', 'intensity', 'error']
    fit_dir_lst = ['fit', 'res', 'weight', 'chi2_ind', 'red_chi2_ind']

    name_of_dset = result['name_of_dset']

    with h5.File(f'{filename}.h5', 'w') as f:

        f.create_dataset('name_of_time_delay_scan_datasets',
                         data=np.char.encode(name_of_dset.astype('str'),
                         encoding='utf-8').astype('S100'),
                         dtype='S100')
        expt = f.create_group('experiment')
        fit_res = f.create_group('fitting_result')
        for i in range(len(name_of_dset)):
            expt_dset = expt.create_group(name_of_dset[i])
            fit_res_dset = fit_res.create_group(name_of_dset[i])
            for k, d in zip(exp_key_lst, exp_dir_lst):
                expt_dset.create_dataset(d, data=result[k][i])
            for k, d in zip(fit_key_lst, fit_dir_lst):
                fit_res_dset.create_dataset(d, data=result[k][i])
            if result['model'] == 'both':
                fit_res_dset.create_dataset(
                    'fit_osc', data=result['fit_osc'][i])
                fit_res_dset.create_dataset(
                    'fit_decay', data=result['fit_decay'][i])
            if result['model'] in ['dmp_osc', 'both']:
                fit_res_dset.create_dataset('phase', data=result['phase'][i])

        for k in model_key_lst:
            fit_res.attrs[k] = result[k]
        fit_res.attrs['same_t0'] = result['same_t0']
        if result['method_glb'] is None:
            fit_res.attrs['method_glb'] = 'no'
        else:
            fit_res.attrs['method_glb'] = result['method_glb']
            fit_res.attrs['message_glb'] = result['message_glb']

        fit_res_param = fit_res.create_group('parameter')
        fit_res_param.create_dataset('param_opt', data=result['x'])
        fit_res_param.create_dataset('param_eps', data=result['x_eps'])
        fit_res_param.create_dataset(
            'param_name',
            data=np.char.encode(result['param_name'].astype('str'),
            encoding='utf-8').astype('S100'), dtype='S100')
        fit_res_param.create_dataset('param_bounds', data=result['bounds'])
        fit_res_param.create_dataset('correlation', data=result['corr'])
        fit_res_mis = fit_res.create_group('miscellaneous')
        fit_res_mis.create_dataset('jac', data=result['jac'])
        fit_res_mis.create_dataset('cov', data=result['cov'])
        fit_res_mis.create_dataset('cov_scaled', data=result['cov_scaled'])


def load_TransientResult(filename: str) -> TransientResult:
    '''
    load transient fitting result from h5 file

    Args:
     filename: filename to load result. It will load filename.h5

    Returns:
     transient fitting result loaded from h5 file
    '''

    model_key_lst = ['model', 'irf', 'fwhm', 'eta', 'n_decay', 'n_osc',
                     'base', 'chi2', 'n_param', 'n_param_ind', 'num_pts', 'red_chi2',
                     'aic', 'bic', 'nfev', 'method_lsq', 'success_lsq', 'message_lsq', 'status']

    exp_key_lst = ['intensity', 'eps']
    fit_key_lst = ['fit', 'res', 'c', 'chi2_ind', 'red_chi2_ind']
    exp_dir_lst = ['intensity', 'error']
    fit_dir_lst = ['fit', 'res', 'weight', 'chi2_ind', 'red_chi2_ind']

    result = TransientResult()

    with h5.File(f'{filename}.h5', 'r') as f:

        expt = f['experiment']
        fit_res = f['fitting_result']

        for k in model_key_lst:
            result[k] = fit_res.attrs[k]

        try:
            result['same_t0'] = fit_res.attrs['same_t0']
        except KeyError:
            result['same_t0'] = False

        if fit_res.attrs['method_glb'] == 'no':
            result['method_glb'] = None
            result['message_glb'] = None
        else:
            result['method_glb'] = fit_res.attrs['method_glb']
            result['message_glb'] = fit_res.attrs['message_glb']

        result['name_of_dset'] = np.char.decode(np.atleast_1d(
            f['name_of_time_delay_scan_datasets']), encoding='utf-8')

        result['t'] = np.empty(len(result['name_of_dset']), dtype=object)

        for k in exp_key_lst:
            result[k] = np.empty(len(result['name_of_dset']), dtype=object)

        for k in fit_key_lst:
            result[k] = np.empty(len(result['name_of_dset']), dtype=object)

        if result['model'] in ['dmp_osc', 'both']:
            result['phase'] = np.empty(
                len(result['name_of_dset']), dtype=object)

        if result['model'] == 'both':
            result['fit_osc'] = np.empty(
                len(result['name_of_dset']), dtype=object)
            result['fit_decay'] = np.empty(
                len(result['name_of_dset']), dtype=object)

        for i in range(len(result['name_of_dset'])):
            expt_dset = expt[result['name_of_dset'][i]]
            fit_res_dset = fit_res[result['name_of_dset'][i]]
            result['t'][i] = np.atleast_1d(expt_dset['time_delay'])
            for k, d in zip(exp_key_lst, exp_dir_lst):
                result[k][i] = np.atleast_2d(expt_dset[d])
            for k, d in zip(fit_key_lst, fit_dir_lst):
                result[k][i] = np.atleast_2d(fit_res_dset[d])

            if result['model'] in ['both', 'dmp_osc']:
                result['phase'][i] = np.atleast_2d(fit_res_dset['phase'])

            if result['model'] == 'both':
                result['fit_osc'][i] = fit_res_dset['fit_osc']
                result['fit_decay'][i] = fit_res_dset['fit_decay']

        fit_res_param = fit_res['parameter']
        result['x'] = np.atleast_1d(fit_res_param['param_opt'])
        result['x_eps'] = np.atleast_1d(fit_res_param['param_eps'])
        result['param_name'] = np.char.decode(np.atleast_1d(fit_res_param['param_name']),
        encoding='utf-8')
        tmp = np.atleast_2d(fit_res_param['param_bounds'])
        lst = tmp.shape[0]*[None]
        for i in range(tmp.shape[0]):
            lst[i] = (tmp[i, 0], tmp[i, 1])
        result['bounds'] = lst
        result['corr'] = np.atleast_2d(fit_res_param['correlation'])

        fit_res_mis = fit_res['miscellaneous']
        result['jac'] = np.atleast_2d(fit_res_mis['jac'])
        result['cov'] = np.atleast_2d(fit_res_mis['cov'])
        result['cov_scaled'] = np.atleast_2d(fit_res_mis['cov_scaled'])

    return result
